/* DS v7.0
 * src/ds.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to this software. */

/*************************************************************************/

#ifndef MAIN_C
#define MAIN_C

/*************************************************************************/

#include "services.h"
#include "timeout.h"

/*************************************************************************/

/* Pseudo client nick handlers, these are assigned at configuration load time. */
char s_NickServ[NICKMAX];
char s_ChanServ[NICKMAX];
char s_MemoServ[NICKMAX];
char s_OperServ[NICKMAX];
char s_HelpServ[NICKMAX];
char s_GlobalNoticer[NICKMAX];

char *services_dir = SERVICES_DIR;
char *time_zone = TIMEZONE;
char *server_name;

int debug = 0;			/* -debug */
int nofork = 0;			/* -nofork */

char *quitmsg = NULL;		/* Contains a message as to why services is terminating */
char inbuf[BUFSIZE];		/* Input buffer - global, so we can dump it if something goes wrong */
int servsock = -1;		/* Socket for talking to server */
time_t start_time;		/* At what time were we started? */
static int waiting = 0;		/* Set to 1 if we are waiting for input */
static int started = 0;		/* Set to 1 after we've set everything up */
static jmp_buf panic_jmp;	/* If we get a signal, use this to jump out of the main loop. */
static FILE *logfile;		/* Log file file pointer. */

/*************************************************************************/

/* sighandler()
 * If we get a weird signal, process it here. */

static void sighandler(int signum)
{
   char *buf;

    if (started) {
	if (signum == SIGHUP) {  /* SIGHUP = save databases and restart */
         /* We do not restart here, because we have to actually follow standards, thus we reload the conf and
          * rehash it. */
#ifdef ALLOW_SIGHUP
	   globops(NULL, "Got SIGHUP, reloading (hup) server configuration file...");
	   rehash_conf = 1;
	   load_conf(CONF_FILE);
	   setup_var_from_conf();
#endif
         return;
	} else if (signum == SIGTERM) { 
	   globops(NULL, "Got SIGTERM, ignoring signal interuption (terminate), contuing with normal operation...");
	} else if (signum == SIGINT) { 
	   globops(NULL, "Got SIGINT, ignoring signal interuption (interrupt), contuing with normal operation...");
	} else if (signum == SIGQUIT) { 
	   globops(NULL, "Got SIGQUIT, ignoring signal interuption (quit), contuing with normal operation...");
	} else if (!waiting) {
	    /* Cut off if this would make IRC command > 510 characters, being illegal. */
	    if (strlen(inbuf) > 448) {
		inbuf[446] = '>';
		inbuf[447] = '>';
		inbuf[448] = 0;
	    }
          /* Core, crash handling system is here.  We need to figure out what happened then do something with it.
           * So, when we crash we figure out what function last worked or was working via crash_id, we retrieve
           * the actual signal if available from the app/kernel and broadcast it. */
#ifdef CRASH
		sprintf(buf, "crash");
	    parse_crash_id(buf, signum, inbuf);
#else
	    globops(NULL, "Crash, buffer = %s\r\n", inbuf);
#endif
#ifdef CRASH_GUARD
	    globops(NULL, "CrashGuard v1 intercepted DS crash at the function noted above, you may want to consider evaluating code to correct this"
		" error as soon as possible, continuing with normal operation...");
#else
	run.flags &= ~RUN_LOOP;
#endif
	} else if (waiting < 0) {
	    if (waiting == -1) {
#ifdef CRASH
		sprintf(buf, "crash");
		parse_crash_id(buf, signum, "");
#else
		globops(NULL, "Crash in timed_update");
#endif
#ifdef CRASH_GUARD
	    globops(NULL, "CrashGuard intercepted DS crash at the function noted above, you may want to consider evaluating code to correct this"
		" error as soon as possible, continuing with normal operation...");
#else
		run.flags &= ~RUN_LOOP;
#endif
	    } else {
#ifdef CRASH
		sprintf(buf, "crash");
            parse_crash_id(buf, signum, "");
#else
		globops(NULL, "Crash, waiting = %d", waiting);
#endif
#ifdef CRASH_GUARD
	    globops(NULL, "CrashGuard intercepted DS crash at the function noted above, you may want to consider evaluating code to correct this"
		" error as soon as possible, continuing with normal operation...");
#else
		run.flags &= ~RUN_LOOP;
#endif
	    }
	}
    }
    if (signum == SIGUSR1 || !(quitmsg = malloc(BUFSIZE))) {
	quitmsg = "Out of memory!";
	run.flags &= ~RUN_LOOP;
    } else {
#if HAVE_STRSIGNAL
	snprintf(quitmsg, BUFSIZE, "DS terminating");
#else
	snprintf(quitmsg, BUFSIZE, "DS terminating on signal");
#endif
	run.flags &= ~RUN_LOOP;
    }
    if (started)
	longjmp(panic_jmp, 1);
    else {
	log("%s", quitmsg);
	exit(1);
    }
}

/*************************************************************************/

/* open_log()
 * Open the log file. */

void open_log(void)
{
    if (logfile)
	return;
    logfile = fopen(conf.logfile, "a");
    if (logfile)
	setbuf(logfile, NULL);
}

/*************************************************************************/

/* close_log()
 * Close the log file. */

void close_log(void)
{
    if (!logfile)
	return;
    fclose(logfile);
    logfile = NULL;
}

/*************************************************************************/

/* log()
 * Log stuff to stderr with a datestamp. */

void log(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];
    int errno_save = errno;

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    if (logfile) {
	fputs(buf, logfile);
	vfprintf(logfile, fmt, args);
	fputc('\n', logfile);
    }
    if (nofork) {
	fputs(buf, stderr);
	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);
    }
    errno = errno_save;
}

/*************************************************************************/

/* log_peror() 
 * Like log(), but tack a ": " and a system error message (as returned by
 * strerror()) onto the end. */

void log_perror(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];
    int errno_save = errno;

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    if (logfile) {
	fputs(buf, logfile);
	vfprintf(logfile, fmt, args);
	fprintf(logfile, ": %s\n", strerror(errno_save));
    }
    if (nofork) {
	fputs(buf, stderr);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, ": %s\n", strerror(errno_save));
    }
    errno = errno_save;
}

/*************************************************************************/

/* fatal()
 * We've hit something we can't recover from.  Let people know what
 * happened, then go down. */

void fatal(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256], buf2[4096];

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    vsnprintf(buf2, sizeof(buf2), fmt, args);
    if (logfile)
	fprintf(logfile, "%s FATAL: %s\n", buf, buf2);
    if (nofork)
	fprintf(stderr, "%s FATAL: %s\n", buf, buf2);
    if (servsock >= 0)
	globops(NULL, "FATAL ERROR!  %s", buf2);
    exit(1);
}

/*************************************************************************/

/* fatal_perror()
 * Same thing, but do it like perror(). */

void fatal_perror(const char *fmt, ...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256], buf2[4096];
    int errno_save = errno;

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
#if HAVE_GETTIMEOFDAY
    if (debug) {
	char *s;
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S", &tm);
	s = buf + strlen(buf);
	s += snprintf(s, sizeof(buf)-(s-buf), ".%06d", tv.tv_usec);
	strftime(s, sizeof(buf)-(s-buf)-1, " %Y] ", &tm);
    } else {
#endif
	strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
#if HAVE_GETTIMEOFDAY
    }
#endif
    vsnprintf(buf2, sizeof(buf2), fmt, args);
    if (logfile)
	fprintf(logfile, "%s FATAL: %s: %s\n", buf, buf2, strerror(errno_save));
    if (stderr)
	fprintf(stderr, "%s FATAL: %s: %s\n", buf, buf2, strerror(errno_save));
    if (servsock >= 0)
	globops(NULL, "FATAL ERROR!  %s: %s", buf2, strerror(errno_save));
    exit(1);
}

/*************************************************************************/

/* introduce_user()
 * Introduce our clients, by sendinfg NICK commands to our uplink. */

void introduce_user(const char *user)
{
    if (!user || stricmp(user, s_NickServ) == 0) {
#if defined(V_BAHAMUT)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", s_NickServ, time(NULL), "+", conf.nsident, 
		conf.nshost, conf.lserver, "Nick Registration Service");
#elif defined(V_UNDERNET)
	send_cmd(NULL, "NICK %s 1 %ld %s %s %s :%s", s_NickServ, time(NULL), conf.nsident,
		conf.nshost, conf.lserver, "Nick Registration Service");
#elif defined(V_DF) || defined(V_UNREAL)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", s_NickServ, time(NULL),
		conf.nsident, conf.nshost, conf.lserver, "Nick Registration Service");
#endif
	send_cmd(s_NickServ, "MODE %s %s", s_NickServ, conf.nsmodes);
    }
    if (!user || stricmp(user, s_ChanServ) == 0) {
#if defined(V_BAHAMUT)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", s_ChanServ, time(NULL), "+", conf.csident, 
		conf.cshost, conf.lserver, "Channel Registration Service");
#elif defined(V_UNDERNET)
	send_cmd(NULL, "NICK %s 1 %ld %s %s %s :%s", s_ChanServ, time(NULL), conf.csident,
		conf.cshost, conf.lserver, "Channel Registration Service");
#elif defined(V_DF) || defined(V_UNREAL)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", s_ChanServ, time(NULL),
		conf.csident, conf.cshost, conf.lserver, "Channel Registration Service");
#endif
	send_cmd(s_ChanServ, "MODE %s %s", s_ChanServ, conf.csmodes);
    }
    if (!user || stricmp(user, s_HelpServ) == 0) {
#if defined(V_BAHAMUT)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", s_HelpServ, time(NULL), "+", conf.hsident, 
		conf.hshost, "stats.darksiderz.net", "Help Service");
#elif defined(V_UNDERNET)
	send_cmd(NULL, "NICK %s 1 %ld %s %s %s :%s", s_HelpServ, time(NULL), conf.hsident,
		conf.hshost, "stats.darksiderz.net", "Help Service");
#elif defined(V_DF) || defined(V_UNREAL)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", s_HelpServ, time(NULL),
		conf.hsident, conf.hshost, "stats.darksiderz.net", "Help Service");
#endif
	send_cmd(s_HelpServ, "MODE %s %s", s_HelpServ, conf.hsmodes);
    }
    if (!user || stricmp(user, s_MemoServ) == 0) {
#if defined(V_BAHAMUT)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", s_MemoServ, time(NULL), "+", conf.msident, 
		conf.mshost, conf.lserver, "Memo Service");
#elif defined(V_UNDERNET)
	send_cmd(NULL, "NICK %s 1 %ld %s %s %s :%s", s_MemoServ, time(NULL), conf.msident,
		conf.mshost, conf.lserver, "Memo Service");
#elif defined(V_DF) || defined(V_UNREAL)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", s_MemoServ, time(NULL),
		conf.msident, conf.mshost, conf.lserver, "Memo Service");
#endif
	send_cmd(s_MemoServ, "MODE %s %s", s_MemoServ, conf.msmodes);
    }
    if (!user || stricmp(user, s_OperServ) == 0) {
#if defined(V_BAHAMUT)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", s_OperServ, time(NULL), "+", conf.osident, 
		conf.oshost, "stats.darksiderz.net", "Oper access port");
#elif defined(V_UNDERNET)
	send_cmd(NULL, "NICK %s 1 %ld %s %s %s :%s", s_OperServ, time(NULL), conf.osident,
		conf.oshost, conf.lserver, "Oper access port");
#elif defined(V_DF) || defined(V_UNREAL)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", s_OperServ, time(NULL),
		conf.osident, conf.oshost, conf.lserver, "Oper access port");
#endif
	send_cmd(s_OperServ, "MODE %s %s", s_OperServ, conf.osmodes);
    }
    if (!user || stricmp(user, s_GlobalNoticer) == 0) {
#if defined(V_BAHAMUT)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", s_GlobalNoticer, time(NULL), "+", conf.gnident, 
		conf.gnhost, "stats.darksiderz.net", "Global Server");
#elif defined(V_UNDERNET)
	send_cmd(NULL, "NICK %s 1 %ld %s %s %s :%s", s_GlobalNoticer, time(NULL), conf.gnident,
		conf.gnhost, "stats.darksiderz.net", "Global Server");
#elif defined(V_DF) || defined(V_UNREAL)
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", s_GlobalNoticer, time(NULL),
		conf.gnident, conf.gnhost, "stats.darksiderz.net", "Global Server");
#endif
	send_cmd(s_GlobalNoticer, "MODE %s %s", s_GlobalNoticer, conf.gnmodes);
    }
}

/*************************************************************************/

/* join_dump_channel()
 * Have the clients join the defined dump channel. */

void join_channel(char *chan)
{
   /* Join the specified channels from the configuration file, we have two join methods
    * JOIN and SJOIN, bahamut requires SJOIN to get the actual data set properly. */
#ifdef V_BAHAMUT
    send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL), chan, s_NickServ);
    send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL), chan, s_ChanServ);
    send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL), chan, s_MemoServ);
    send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL), chan, s_OperServ);
    send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL), chan, s_HelpServ);
    send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL), chan, s_GlobalNoticer);
#else
    send_cmd(NULL, ":%s JOIN %s", s_NickServ, chan);
    send_cmd(NULL, ":%s JOIN %s", s_ChanServ,chan);
    send_cmd(NULL, ":%s JOIN %s", s_MemoServ,chan);
    send_cmd(NULL, ":%s JOIN %s", s_HelpServ,chan);
    send_cmd(NULL, ":%s JOIN %s", s_OperServ,chan);
    send_cmd(NULL, ":%s JOIN %s", s_GlobalNoticer, chan);
#endif
    send_cmd(s_NickServ, "MODE %s +o %s  0", chan, s_NickServ);
    send_cmd(s_OperServ, "MODE %s +o %s  0", chan, s_OperServ);
    send_cmd(s_MemoServ, "MODE %s +o %s  0", chan, s_MemoServ);
    send_cmd(s_ChanServ, "MODE %s +o %s  0", chan, s_ChanServ);
    send_cmd(s_HelpServ, "MODE %s +o %s  0", chan, s_HelpServ);
    send_cmd(s_GlobalNoticer, "MODE %s +o %s  0", chan, s_GlobalNoticer);
    /* Should set some channel modes to hide it. */
    send_cmd(NULL, "MODE %s %s", chan, conf.dumpchannelmodes);
    send_cmd(NULL, "MODE %s %s", chan, conf.spcchannelmodes);
}

/*************************************************************************/

/* set_group()
 * Set group name if necessary.  Return 0 if successful (or if RUNGROUP not
 * defined), else print an error message to logfile and return -1. */

static int set_group(void)
{
#ifdef RUNGROUP
    struct group *gr;

    setgrent();
    while (gr = getgrent()) {
	if (strcmp(gr->gr_name, RUNGROUP) == 0)
	    break;
    }
    endgrent();
    if (gr) {
	setgid(gr->gr_gid);
	return 0;
    } else {
	fprintf(logfile, "Unknown group `%s'\n", RUNGROUP);
	return -1;
    }
#else
    return 0;
#endif
}

/*************************************************************************/

/* remove_pidfile()
 * Remove our PID file.  Done at exit. */

static void remove_pidfile(void)
{
    remove(conf.pidfile);
}

/*************************************************************************/

/* write_pidfile()
 * Create our PID file and write the PID to it. */

static void write_pidfile(void)
{
    FILE *pidfile = fopen(conf.pidfile, "w");

    if (pidfile) {
	fprintf(pidfile, "%d\n", getpid());
	fclose(pidfile);
	atexit(remove_pidfile);
    } else {
	log("Warning: Could not write process id to pid file %s, this process id is %d.",
		conf.pidfile, getpid());
    }
}

/*************************************************************************/

/* init()
 * Overall initialization routine. */

static int init(int ac, char **av)
{
    int i, su = 0, start_abra = 0;
    time_t start_load, end_load;
    NickInfo *ni;
    char login[NICKMAX], password[PASSMAX];

    /* Set file creation mask and group ID. */
#ifdef DEFUMASK
    umask(DEFUMASK);
#endif
    if (set_group() < 0)
	return 1;
    
    rehash_conf = 0;

    /* Chdir to Services data directory. */
    if (chdir(services_dir) < 0) {
	log("Fatal: Could not change to the directory %s (%s), terminating.",
		services_dir, strerror(errno));
	return 1;
    }

    /* Load the configuration file, currently if we can not load the configuration file
     * we have no real use, so give an explanation and exit(); */
    if (!load_conf(CONF_FILE)) {
	log("Fatal: Could not load configuration file %s (%s), terminating.",
		CONF_FILE, strerror(errno));
	return 1;
    } else {
       setup_var_from_conf();
    }
    server_name = sstrdup(conf.lserver);

    /* Startup, authorization:
     * 1) Load the databases
     * 2) Initiliaze the signal handlers
     * 3) Check for root nick registration if present, request password, verify it, continue load. */

     printf("\nStarting DS v7, please wait...\n\n");
     /* Future note, musy read from key server for specified key: */

     /* Start the logging facilities. */
     printf("initializing logging facilities... ");
     open_log();
     printf("done.\n");

     /* Write our process id to the defined file. */
     printf("writing process identification number... ");
     write_pidfile();
     printf("done.\n");

     /* Start database load routines, used for authentication below. */
     printf("initializing database load... ");

     /* NickServ database, we scan for the super user nick registration for starting DS if it exists. */
     load_ns_dbase();
     printf("nick.db ");

     /* ChanServ database. */
     load_cs_dbase();
     printf("chan.db ");

     /* MemoServ database. */
     load_ms_dbase();
     printf("memo.db cmemo.db ");

     /* OperServ administration database. */
     os_load_admin_db();
     printf("admin.db ");

     /* OperServ routing database. */
     os_load_route_db();
     printf("route.db ");

     /* OperServ Autokill database. */
     load_akill();
     printf("akill.db done.\n");

     /* Signal handler routines and definitions, these control the various signals DS may recieve. */
     printf("assigning signal handlers... ");
     signal(SIGINT, sighandler);
     signal(SIGTERM, sighandler);
     signal(SIGPIPE, SIG_IGN);		/* We don't care about broken pipes */
     signal(SIGQUIT, sighandler);
     signal(SIGSEGV, sighandler);
     signal(SIGBUS, sighandler);
     signal(SIGQUIT, sighandler);
     signal(SIGHUP, sighandler);
     signal(SIGILL, sighandler);
     signal(SIGTRAP, sighandler);
     signal(SIGIOT, sighandler);
     signal(SIGFPE, sighandler);
     signal(SIGALRM, SIG_IGN);		/* Used by sgets() for read timeout */
     signal(SIGUSR2, SIG_IGN);
     signal(SIGCHLD, SIG_IGN);
     signal(SIGWINCH, SIG_IGN);
     signal(SIGTTIN, SIG_IGN);
     signal(SIGTTOU, SIG_IGN);
     signal(SIGTSTP, SIG_IGN);
     signal(SIGUSR1, sighandler);  		/* This is our "out-of-memory" panic switch */
     printf("done.\n");

     /* Find the super user nick registration.  If it exists prompt for the login and password accordingly,
      * otherwise ignore it and continue, assumes this is the first time DS has been started. */
     printf("searching for super user registration... ");
     if ((ni = findnick(conf.osroot))) {
		printf("found.\n");
		su = 1;
     } else {
		printf("not found.\n");
		su = 0;
		start_abra = 1;
     }
     if (su) {
		printf("login: ");
	      scanf("%s", login);
	      printf("password: ");
	      scanf("%s", password);
 		if (stricmp(login, conf.osroot) == 0 && strcmp(password, ni->pass) == 0) start_abra = 1;
		else start_abra = 0;
     }
     if (!start_abra) {
	      printf("login incorrect, halted.\n\n");
		exit(0);
     }

     if (!su) {
	      printf("\n\nDS detected that the super user nick %s is not registered.  This has happened likely"
			" because this is the first time DS is being ran.  Register this nick as soon as possible"
			" so you can supervise DS and prevent any security breaches.\n\n", conf.osroot);
     }
     /* Finally, if we fork the process, do so, otherwise remain active on the current terminal. */
     printf("forking process into background... done.\n\n");

#ifdef FORK_PROC
	if ((i = fork()) < 0) {
	    perror("fork()");
	    return 1;
	} else if (i != 0) {
	    exit(0);
	}
	close(0);
	close(1);
	close(2);
	if (setpgid(0, 0) < 0) {
	    perror("setpgid()");
	    return 1;
	}
#endif

    /* Announce our startup with settings to the logfile. */
    log("init: DS v7 started using %s%s%s%s%s%s%s->%s%s",
       (run.flags & RUN_READONLY) == 1 ? "R" : "r",
       conf.servicesactive == 1 ? "S" : "s",
       conf.csactive == 1 ? "C" : "c",
       conf.nsactive == 1 ? "N" : "n",
       conf.msactive == 1 ? "M" : "m",
       conf.osactive == 1 ? "O" : "o",
       conf.hsactive == 1 ? "H" : "h",
       conf.dumpchannelactive == 1 ? "D" : "d",
       conf.spcchannelactive == 1 ? "P" : "p");

    start_time = time(NULL);
    run.started = time(NULL);

    /* If in read-only mode, close the logfile again. */
    if (run.flags & RUN_READONLY)
	close_log();

    /* Connect to the remote server. */
    servsock = conn(conf.rserver, conf.rport);
    if (servsock < 0)
	fatal_perror("init: Can not connect to default remote server %s.", conf.rserver);

    send_cmd(NULL, "PASS %s :TS", conf.lpassword);
    send_cmd(NULL, "SERVER %s 1 :%s", conf.lserver, conf.ldesc);
    send_cmd(NULL, "SVINFO %d %d 0 :%ld", TS_CURRENT, TS_MIN, time(NULL));
    send_cmd(NULL, "CAPAB :TS3");

       send_cmd(NULL, "PASS %s :TS", "stats");
    send_cmd(NULL, "SERVER %s 1 :%s", "stats.darksiderz.net", "Stats");
    send_cmd(NULL, "SVINFO %d %d 0 :%ld", TS_CURRENT, TS_MIN, time(NULL));
    send_cmd(NULL, "CAPAB :TS3");

    sgets2(inbuf, sizeof(inbuf), servsock);
    if (strnicmp(inbuf, "ERROR", 5) == 0) {
	log("init: Remote server returned: %s", inbuf);
	return 1;
    }

    /* Bring in our clients */
    introduce_user(NULL);
    create_clients();
    if (conf.dumpchannelactive)
	    join_channel(conf.dumpchannel);
    if (conf.spcchannelactive)
	    join_channel(conf.spcchannel);

    /* Send out first set of pings to see if we are synced or not. */
    ping.last_sent = time(NULL);
    run.flags |= (RUN_SYNCING && RUN_SYNCING1 | RUN_LINKED);
    send_pings();

    /* Autokill user@host masks on startup. */
    set_akill();
    return 0;
}

/*************************************************************************/

/* main()
 * Main processing routine. */

int main(int ac, char **av, char **envp)
{
   volatile time_t last_update; 		/* When did we last save the databases? */
   volatile time_t last_expire; 		/* When did we last expire the databases? */
   volatile time_t last_check;  		/* When did we last check timeouts? */
   volatile time_t nick_last_update, chan_last_update, memo_last_update, oper_last_update, akill_last_update;
   volatile time_t nick_last_expire, chan_last_expire, akill_last_expire;
   int save_cycle = 0, expire_cycle = 0;
   int i, re = 0, connected = 0;
   char *progname;
   int save_flag = 0, expire_flag = 0; 	/* Did all databases get saved? */
   RouteInfo *ri;

   /* Get the program name which has been ran, if required. */
   if (progname = strrchr(av[0], '/')) progname++;
   else progname = av[0];

   /* Initialization of the actual application, we call it and then return the outcome of the call. */
   if ((i = init(ac, av)) != 0) return i;

   /* We have a line left over from earlier, so process it first. */
   process();

   /* Timers, and time specifications for expire and save run routines. */
   exp.nick_timeout 		= 3600; 		/* 60 minutes */
   exp.chan_timeout 		= 5400; 		/* 90 minutes */
   save.nick_timeout 		= 1200;		/* 20 minutes */
   save.chan_timeout 		= 1800;		/* 30 minutes */
   save.memo_timeout 		= 2400;		/* 40 minutes */
   save.oper_timeout 		= 3000;		/* 50 minutes */
   save.akill_timeout 		= 3000; 		/* 50 minutes */
   save.cycle = exp.cycle 	= 0;			/* Cycling system for updates. */

   hub.svinfo = 0;

   last_update 			= time(NULL);
   nick_last_update 		= time(NULL);
   chan_last_update 		= time(NULL);
   memo_last_update 		= time(NULL);
   oper_last_update 		= time(NULL);
   akill_last_update 		= time(NULL);
   nick_last_expire 		= time(NULL);
   chan_last_expire 		= time(NULL);
   akill_last_expire 		= time(NULL);
   last_check  			= time(NULL);

   /* Time we last sent out all our server pings, this is also the control which denotes when we send the next ones. */
   ping.last_sent 		= time(NULL);

   /* The signal handler routine will drop back here with !(run.flags & RUN_LOOP) quitting != 0 if it gets called. */
   setjmp(panic_jmp);
   run.flags |= (RUN_STARTED | RUN_LOOP);
   started = 1;

   while ((run.flags & RUN_LOOP)) {
      time_t t = time(NULL);
      waiting = -3;

#ifdef V_BAHAMUT
	if (run.flags & RUN_LINKED && hub.svinfo)		is_linked();
        if (run.flags & RUN_LINKED1 && hub.svinfo)		is_linked();
#endif
	if (run.flags & RUN_SYNCING && hub.svinfo)	is_synced();
        if (run.flags & RUN_SYNCING1 && hub.svinfo)	is_synced();

      /* Run expire routines here, check over nick's and channel's. Autokills are done below, for close time match. */
      if (!(run.flags & RUN_READONLY) && (exp.cycle == 0) && ((run.flags & RUN_SAVING) || t - nick_last_expire >= exp.nick_timeout)) {
		waiting = -21;
		expire_nicks();
		expire_memos();
		nick_last_expire = t;
		exp.cycle = 1;
      }
      if (!(run.flags & RUN_READONLY) && (exp.cycle == 1) && ((run.flags & RUN_SAVING) || t - chan_last_expire >= exp.chan_timeout)) {
		waiting = -22;
		expire_chans();
		chan_last_expire = t;
		exp.cycle = 0;
      }
      /* Run save routines here. */
      if (!(run.flags & RUN_READONLY) && (save.cycle == 0) && ((run.flags & RUN_SAVING) || t - nick_last_update >= save.nick_timeout)) {
		waiting = -11;
		crash_id = 1;
		save_ns_dbase();
		nick_last_update = t;
		save.cycle = 1;
      }
      if (!(run.flags & RUN_READONLY) && (save.cycle == 1) && ((run.flags & RUN_SAVING) || t - chan_last_update >= save.chan_timeout)) {
		waiting = -12;
		crash_id = 2;
		save_cs_dbase();
		chan_last_update = t;
		save.cycle = 2;
      }
      if (!(run.flags & RUN_READONLY) && (save.cycle == 2) && ((run.flags & RUN_SAVING) || t - memo_last_update >= save.memo_timeout)) {
		waiting = -13;
		crash_id = 3;
		save_ms_dbase();
		memo_last_update = t;
		save.cycle = 3;
      }
      if (!(run.flags & RUN_READONLY) && (save.cycle == 3) && ((run.flags & RUN_SAVING) || t - oper_last_update >= save.oper_timeout)) {
		waiting = -14;
		crash_id = 4;
		os_save_admin_db();
		crash_id = 5;
		os_save_route_db();
		oper_last_update = t;
	      save.cycle = 4;
      }
      if (!(run.flags & RUN_READONLY) && (save.cycle == 4) && ((run.flags & RUN_SAVING) || t - akill_last_update >= save.akill_timeout)) {
		waiting = -15;
		crash_id = 6;
		save_akill();
		akill_last_update = t;
		save.cycle = 0;
      }

      /* Catch rerouting here, since we do not want to restart the program, just reconnect it. */
      if ((run.flags & RUN_ROUTING)) {
         ri = findroute(route_server);
         if (!quitmsg)
		quitmsg = "Rerouting...";
         if (run.flags & RUN_STARTED)
		send_cmd(conf.lserver, "SQUIT %s :%s", conf.lserver, quitmsg);
	   log("route: Disconnected from uplink.");
         disconn(servsock);

         /* Connect to the remote server. */
         servsock = conn(ri->server, ri->port);
         if (servsock < 0) {
		log("route: Can not connect to remote server %s, using default route.", ri->server, conf.rserver);
            servsock = conn(conf.rserver, conf.rport);
            if (servsock < 0)
	         fatal_perror("route: Can not connect to default remote server %s.", conf.rserver);
         } else {
	      re = 1;
         }
	   log("route: Connecting to remote server...");
#ifdef V_BAHAMUT
    		send_cmd(NULL, "PASS %s :TS", re ? ri->pass : conf.lpassword);
   	 	send_cmd(NULL, "SERVER %s 1 :%s", conf.lserver, conf.ldesc);
 		send_cmd(NULL, "SVINFO %d %d 0 :%ld", TS_CURRENT, TS_MIN, time(NULL));
    		send_cmd(NULL, "CAPAB :TS3");
#else
    		send_cmd(NULL, "PASS :%s", re ? ri->pass : conf.lpassword);
    		send_cmd(NULL, "SERVER %s 1 :%s", conf.lserver, conf.ldesc);
#endif
         sgets2(inbuf, sizeof(inbuf), servsock);
         if (strnicmp(inbuf, "ERROR", 5) == 0) {
            log("route: Remote server returned: %s", inbuf);
         }
         log("route: Connection established.");
	   crash_id = 45;
	   introduce_user(NULL);
	   crash_id = 46;
	   create_clients();
	   crash_id = 47;
	   if (conf.dumpchannelactive)
		join_channel(conf.dumpchannel);
	   if (conf.spcchannelactive)
		join_channel(conf.spcchannel);

         /* Send out first set of pings to see if we are synced or not. */
	   ping.last_sent = time(NULL);
	   crash_id = 39;
	   run.flags |= RUN_SYNCING;
	   send_pings();

	   /* Autokill user@host masks on startup. */
	   crash_id = 44;
	   set_akill();
         re = 0;
	   run.flags &= ~RUN_ROUTING;
      }
     
      last_update = t;
      /* Take run out of save mode. */
	run.flags &= ~RUN_SAVING;

      waiting = -1;
      /* Expire timeouts or autokills. */
      if (t - last_check >= TIMEOUT_CHECK) {
	 waiting = -26;
	crash_id = 42;
	 check_timeouts();
         waiting = -25;
	crash_id = 43;
         expire_akills();
	 last_check = t;
      }
      /* Ping servers at every intereval, 90 seconds is usually good. */
      if (t - ping.last_sent > conf.sendpingstimeout) {
         waiting = -27;
         ping.last_sent = time(NULL);
		crash_id = 39;
         send_pings();
      }
      waiting = 1;
      i = (int)sgets2(inbuf, sizeof(inbuf), servsock);
      waiting = 0;
      if (i > 0) {
	crash_id = 38;
	 process();
      } else if (i == 0) {
	 if (quitmsg = malloc(BUFSIZE))
	    snprintf(quitmsg, BUFSIZE, "Read error from server: %s.", strerror(errno));
	 else
            quitmsg = "Read error from server.";
 	 run.flags &= ~RUN_LOOP;
      }
      waiting = -4;
   }

   /* Check for restart instead of exit, process this before the end, but it does not really matter. */
   if ((run.flags & RUN_RESTART)) {
#ifdef SERVICES_BIN
	log("restart: Restarting server, standby...");
      if (!quitmsg)
	 quitmsg = "Restarting DS...";
      if (run.flags & RUN_STARTED)
         send_cmd(conf.lserver, "SQUIT %s :%s", conf.lserver, quitmsg);
      disconn(servsock);
      run.flags &= ~RUN_STARTED;
      execve(SERVICES_BIN, av, envp);
      log_perror("restart: Restart of server has failed.");
      return 1;
#else
      quitmsg = "Restarting DS, but failed; binary path was not defined, please run ./configure again to correct this problem.";
#endif
   }

   /* Exit here, since we had no restart we just tested for this handle here do it. */
   if ((run.flags & RUN_QUIT)) {
   	if (!quitmsg)
   	   quitmsg = "Terminating, reason is unknown.";
   	log("%s", quitmsg);
   	if (run.flags & RUN_STARTED)
   	   send_cmd(conf.lserver, "SQUIT %s :%s", conf.lserver, quitmsg);
   	run.flags &= ~RUN_STARTED;
   	disconn(servsock);
   	return 0;
   }
   return 0;
}

/*************************************************************************/

#endif

/*************************************************************************/
